#include <konkret/konkret.h>
#include "LMI_IPNetworkConnectionCapabilities.h"
#include <LMI_IPAssignmentSettingData.h>
#include "globals.h"
#include "network.h"
#include "port.h"
#include "connection.h"
#include "setting.h"

static const CMPIBroker* _cb = NULL;

static void LMI_IPNetworkConnectionCapabilitiesInitialize(CMPIInstanceMI *mi)
{
    mi->hdl = network_ref(_cb);
}

static CMPIStatus LMI_IPNetworkConnectionCapabilitiesCleanup(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    CMPIBoolean term)
{
    network_unref(mi->hdl);
    mi->hdl = NULL;
    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_IPNetworkConnectionCapabilitiesEnumInstanceNames(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop)
{
    return KDefaultEnumerateInstanceNames(
        _cb, mi, cc, cr, cop);
}

static CMPIStatus LMI_IPNetworkConnectionCapabilitiesEnumInstances(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char** properties)
{
    CMPIStatus res = { CMPI_RC_OK, NULL };
    Network *network = mi->hdl;
    const char *ns = KNameSpace(cop);
    char *instanceid;

    Port *port;
    network_lock(network);
    const Ports *ports = network_get_ports(network);
    for (size_t i = 0; i < ports_length(ports); ++i) {
        port = ports_index(ports, i);

        LMI_IPNetworkConnectionCapabilities w;
        LMI_IPNetworkConnectionCapabilities_Init(&w, _cb, ns);

        instanceid = id_to_instanceid(port_get_id(port),
                "LMI_IPNetworkConnectionCapabilities");
        LMI_IPNetworkConnectionCapabilities_Set_InstanceID(&w, instanceid);
        free(instanceid);

        if (!ReturnInstance(cr, w)) {
            error("Unable to return instance of class "
                    LMI_IPNetworkConnectionCapabilities_ClassName);
            CMSetStatus(&res, CMPI_RC_ERR_FAILED);
            break;
        }
    }
    network_unlock(network);
    return res;
}

static CMPIStatus LMI_IPNetworkConnectionCapabilitiesGetInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char** properties)
{
    return KDefaultGetInstance(
        _cb, mi, cc, cr, cop, properties);
}

static CMPIStatus LMI_IPNetworkConnectionCapabilitiesCreateInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const CMPIInstance* ci)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_IPNetworkConnectionCapabilitiesModifyInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const CMPIInstance* ci,
    const char** properties)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_IPNetworkConnectionCapabilitiesDeleteInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_IPNetworkConnectionCapabilitiesExecQuery(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* lang,
    const char* query)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

CMInstanceMIStub(
    LMI_IPNetworkConnectionCapabilities,
    LMI_IPNetworkConnectionCapabilities,
    _cb,
    LMI_IPNetworkConnectionCapabilitiesInitialize(&mi))

static void LMI_IPNetworkConnectionCapabilitiesMethodInitialize(CMPIMethodMI *mi)
{
    mi->hdl = network_ref(_cb);
}

static CMPIStatus LMI_IPNetworkConnectionCapabilitiesMethodCleanup(
    CMPIMethodMI* mi,
    const CMPIContext* cc,
    CMPIBoolean term)
{
    network_unref(mi->hdl);
    mi->hdl = NULL;
    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_IPNetworkConnectionCapabilitiesInvokeMethod(
    CMPIMethodMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* meth,
    const CMPIArgs* in,
    CMPIArgs* out)
{
    return LMI_IPNetworkConnectionCapabilities_DispatchMethod(
        _cb, mi, cc, cr, cop, meth, in, out);
}

CMMethodMIStub(
    LMI_IPNetworkConnectionCapabilities,
    LMI_IPNetworkConnectionCapabilities,
    _cb,
    LMI_IPNetworkConnectionCapabilitiesMethodInitialize(&mi))

KUint16 LMI_IPNetworkConnectionCapabilities_CreateGoalSettings(
    const CMPIBroker* cb,
    CMPIMethodMI* mi,
    const CMPIContext* context,
    const LMI_IPNetworkConnectionCapabilitiesRef* self,
    const KStringA* TemplateGoalSettings,
    KStringA* SupportedGoalSettings,
    CMPIStatus* status)
{
    KUint16 result = KUINT16_INIT;

    KSetStatus(status, ERR_NOT_SUPPORTED);
    return result;
}

static int get_bond_id(Network *network)
{
    const Ports *ports = network_get_ports(network);
    const char *name;
    int max = -1, id;
    for (size_t i = 0; i < ports_length(ports); ++i) {
        name = port_get_id(ports_index(ports, i));
        if (strncmp(name, "bond", 4) == 0) {
            id = atoi(name + 4);
            if (id > max) {
                max = id;
            }
        }
    }
    return max;
}

KUint16 LMI_IPNetworkConnectionCapabilities_LMI_CreateIPSetting(
    const CMPIBroker* cb,
    CMPIMethodMI* mi,
    const CMPIContext* context,
    const LMI_IPNetworkConnectionCapabilitiesRef* self,
    const KString* Caption,
    const KUint16* Type,
    const KUint16* IPv4Type,
    const KUint16* IPv6Type,
    KRef* SettingData,
    CMPIStatus* status)
{
    debug("CreateIPSetting(Caption=\"%s\", Type=%d, IPv4Type=%d, IPv6Type=%d)",
          Caption->chars, Type->value, IPv4Type->value, IPv6Type->value);
    Network *network = mi->hdl;

    KUint16 result = KUINT16_INIT;

    int type = CONNECTION_TYPE_ETHERNET;
    if (Type->exists && !Type->null) {
        if (Type->value != CONNECTION_TYPE_ETHERNET &&
            Type->value != CONNECTION_TYPE_BOND &&
            Type->value != CONNECTION_TYPE_BRIDGE) {

            error("Type parameter has invalid value %d", Type->value);
            KSetStatus2(_cb, status, ERR_INVALID_PARAMETER, "Type parameter has invalid value");
            return result;
        } else {
            type = Type->value;
        }
    }

    int ipv4type = SETTING_METHOD_DISABLED;
    if (IPv4Type->exists && !IPv4Type->null) {
        if (IPv4Type->value != SETTING_METHOD_DISABLED &&
            IPv4Type->value != SETTING_METHOD_STATIC &&
            IPv4Type->value != SETTING_METHOD_DHCP) {

            error("IPv4Type parameter has invalid value %d", IPv4Type->value);
            KSetStatus2(_cb, status, ERR_INVALID_PARAMETER, "IPv4Type parameter has invalid value");
            return result;
        } else {
            ipv4type = IPv4Type->value;
        }
    }

    int ipv6type = SETTING_METHOD_DISABLED;
    if (IPv6Type->exists && !IPv6Type->null) {
        if (IPv6Type->value != SETTING_METHOD_DISABLED &&
            IPv6Type->value != SETTING_METHOD_STATIC &&
            IPv6Type->value != SETTING_METHOD_DHCPv6 &&
            IPv6Type->value != SETTING_METHOD_STATELESS) {

            error("IPv6Type parameter has invalid value %d", IPv6Type->value);
            KSetStatus2(_cb, status, ERR_INVALID_PARAMETER, "IPv6Type parameter has invalid value");
            return result;
        } else {
            ipv6type = IPv6Type->value;
        }
    }

    if (ipv4type == SETTING_METHOD_DISABLED && ipv6type == SETTING_METHOD_DISABLED) {
        error("Can't create SettingData with both IPv4 and IPv6 disabled");
        KSetStatus2(_cb, status, ERR_INVALID_PARAMETER, "Can't create SettingData with both IPv4 and IPv6 disabled");
        return result;
    }

    const char *id = id_from_instanceid(self->InstanceID.chars, "LMI_IPNetworkConnectionCapabilities");

    network_lock(network);
    const Ports *ports = network_get_ports(network);
    Port *port = NULL;
    for (size_t i = 0; i < ports_length(ports); ++i) {
        if (strcmp(port_get_id(ports_index(ports, i)), id) == 0) {
            port = ports_index(ports, i);
            break;
        }
    }

    const char *caption = NULL;
    if (Caption->exists && !Caption->null) {
        caption = Caption->chars;
    }

    Connection *connection = connection_new(network, NULL, caption);
    connection_set_port(connection, port);

    connection_set_type(connection, type);
    Setting *setting;
    char *name;
    switch (type) {
        case CONNECTION_TYPE_BOND:
            setting = setting_new(SETTING_TYPE_BOND);
            asprintf(&name, "bond%d", get_bond_id(network) + 1);
            setting_set_option(setting, "interface-name", name);
            free(name);
            connection_add_setting(connection, setting);
            break;

        case CONNECTION_TYPE_BRIDGE:
            setting = setting_new(SETTING_TYPE_BRIDGE);
            // TODO: add the bridge support
            connection_add_setting(connection, setting);
            break;
    }

    // IPv4
    setting = setting_new(SETTING_TYPE_IPv4);
    setting_set_method(setting, ipv4type);
    if (ipv4type == SETTING_METHOD_STATIC) {
        // Static connection needs to have at least one address
        setting_add_ip_address(setting, SETTING_METHOD_STATIC, "0.0.0.1", 24, NULL);
    }
    connection_add_setting(connection, setting);

    // IPv6
    setting = setting_new(SETTING_TYPE_IPv6);
    setting_set_method(setting, ipv6type);
    if (ipv6type == SETTING_METHOD_STATIC) {
        // Static connection needs to have at least one address
        setting_add_ip_address(setting, SETTING_METHOD_STATIC, "1::1", 24, NULL);
    }
    connection_add_setting(connection, setting);

    int res = network_create_connection(network, connection);
    if (res == 0) {
        // Return reference to LMI_IPAssignmentSettingData
        const char *ns = KNameSpace(LMI_IPNetworkConnectionCapabilitiesRef_ToObjectPath(self, NULL));
        LMI_IPAssignmentSettingDataRef ref;
        LMI_IPAssignmentSettingDataRef_Init(&ref, _cb, ns);
        char *instanceid = id_to_instanceid(connection_get_id(connection), "LMI_IPAssignmentSettingData");
        LMI_IPAssignmentSettingDataRef_Set_InstanceID(&ref, instanceid);
        free(instanceid);

        KRef_SetObjectPath(SettingData, LMI_IPAssignmentSettingDataRef_ToObjectPath(&ref, NULL));
    }

    connection_free(connection);
    network_unlock(network);
    if (res == 0) {
        KSetStatus(status, OK);
    }
    KUint16_Set(&result, res);
    return result;
}

KONKRET_REGISTRATION(
    "root/cimv2",
    "LMI_IPNetworkConnectionCapabilities",
    "LMI_IPNetworkConnectionCapabilities",
    "instance method")
